@building-option +create-set-co
@create $utils named Image Utilities:Image Utilities,imgu,image_utils
;_.("aliases") = {"Image Utilities", "imgu", "image_utils"}
;_.("unique") = 0

@verb _:"jpeg_size" this none this
@program _:jpeg_size
"Pass a JPEG file that has been through decode_binary(var, 1) to this verb to get the size of the JPEG.";
":jpeg_size(array_of_bytes) => {height, width}";
{image} = args;
if (image[1..2] != {255, 216})
  raise(E_INVARG, "Not a JPEG file");
endif
loc = 3;
while (1)
  discarded = 0;
  while (image[loc] != 255)
    loc = loc + 1;
    discarded = discarded + 1;
  endwhile
  while (image[loc] == 255)
    loc = loc + 1;
  endwhile
  if (discarded != 0)
    raise(E_INVARG, "No marker");
  endif
  if (image[loc] in {192, 193, 194, 195, 197, 198, 199, 201, 202, 203, 205, 206, 207})
    "Image length marker";
    loc = loc + 4;
    height = (image[loc] * 255) + image[loc + 1];
    loc = loc + 2;
    width = (image[loc] * 255) + image[loc + 1];
    loc = loc + 3;
    return {height, width};
  elseif (image[loc] in {217, 218})
    raise(E_INVARG, "D9/DA type encountered in JPEG");
  else
    len = (image[loc + 1] * 256) + image[loc + 2];
    loc = (loc + len) + 1;
  endif
endwhile
.

@verb _:"gif_size" this none this
@program _:gif_size
{image, ?behavior = 0} = args;
len = length(image);
if ((len < 15) || ((image[1..6] != {71, 73, 70, 56, 57, 97}) && (image[1..6] != {71, 73, 70, 56, 59, 97})))
  raise(E_INVARG, "Invalid GIF data");
endif
loc = 7;
if (behavior == 0)
  return {image[loc] + (image[loc + 1] * 255), image[loc + 2] + (image[loc + 3] * 255)};
endif
bits = $local.crypto:BLFromFloat(tofloat(image[loc + 4]));
cmapflag = bits[32 - 7];
if (cmapflag)
  skip = (bits[32] + (bits[31] * 2)) + (bits[30] * 4);
  loc = loc + (3 * (2 ^ (((bits[32] + (bits[31] * 2)) + (bits[30] * 4)) + 1)));
endif
loc = loc + 7;
width = height = 0;
while (loc < len)
  if (image[loc] == 44)
    if ((loc + 7) > len)
      raise(E_INVARG, "Corrupted GIF");
    endif
    x = image[loc + 5] + (image[loc + 6] * 255);
    y = image[loc + 7] + (image[loc + 8] * 255);
    if (behavior == 1)
      return {x, y};
    endif
    if ((x > width) && (y > height))
      width = x;
      height = y;
    endif
    loc = loc + 9;
  elseif (image[loc] == 33)
    loc = loc + 1;
    if (image[loc] == 249)
      loc = loc + 7;
    elseif (image[loc] == 254)
      while ((loc < len) && (image[loc] != 0))
        loc = loc + 1;
      endwhile
    elseif (image[loc] == 1)
      loc = loc + 14;
      while ((loc < len) && (image[loc] != 0))
        loc = loc + 1;
      endwhile
    elseif (image[loc] == 255)
      loc = loc + 13;
      while ((loc < len) && (image[loc] != 0))
        loc = loc + 1;
      endwhile
    else
      raise(E_INVARG, tostr("Invalid GIF data, unknown extension ", image[loc]));
    endif
  else
    raise(E_INVARG, tostr("Invalid GIF data, unknown code ", image[loc]));
  endif
endwhile
.

@verb _:"image_size" this none this
@program _:image_size
{image} = args;
len = length(image);
if ((len > 2) && (image[1..2] == {255, 216}))
  return {@this:jpeg_size(image), "image/jpeg"};
elseif ((len > 15) && ((image[1..6] == {71, 73, 70, 56, 57, 97}) || (image[1..6] != {71, 73, 70, 56, 59, 97})))
  return {@this:gif_size(image), "image/gif"};
else
  raise(E_INVARG, "Unsupported image type");
endif
.

@verb _:"base64_encode" this none this
@program _:base64_encode
{s} = args;
alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
s = {@decode_binary(s, 1), 0, 0, 0};
len = length(s) - 3;
i = 1;
result = "";
while (i <= len)
  {a, b, c} = s[i..i + 2];
  result = (((result + alphabet[(a / 4) + 1]) + alphabet[(((a - ((a / 4) * 4)) * 16) + (b / 16)) + 1]) + alphabet[(((b - ((b / 16) * 16)) * 4) + (c / 64)) + 1]) + alphabet[(c - ((c / 64) * 64)) + 1];
  i = i + 3;
endwhile
len = len % 3;
if (len == 1)
  result[$ - 1..$] = "==";
elseif (len == 2)
  result[$] = "=";
endif
return result;
"Copied from Ben (#269):b64e Tue Aug 21 13:45:16 2007 PDT";
.

@verb _:"base64_decode" this none this
@program _:base64_decode
{s} = args;
"leaving A off of the alphabet makes index(alphabet, \"A\") => 0, which is right";
alphabet = "BCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
len = (length(s) / 4) * 3;
if (s[$ - 1..$] == "==")
  len = len - 2;
elseif (s[$] == "=")
  len = len - 1;
endif
result = {};
i = 1;
end = length(s);
while (i < end)
  a = index(alphabet, s[i], 1);
  b = index(alphabet, s[i + 1], 1);
  c = index(alphabet, s[i + 2], 1);
  d = index(alphabet, s[i + 3], 1);
  result = {@result, (a * 4) + (b / 16), ((b % 16) * 16) + (c / 4), ((c % 4) * 64) + d};
  i = i + 4;
endwhile
return encode_binary(result[1..len]);
"Copied from Ben (#269):b64d Tue Aug 21 13:45:23 2007 PDT";
.

"***finished***
