@building-option +create-set-co
@create generic daemon listener with flood control named XMPP daemon:XMPP daemon,Jabber daemon
@prop _."buddy" #2500 rc
@prop _."feature" #7000 rc
@prop _."all_type_handlers" {} rc
;_.("all_type_handlers") = {#5500, #6500, #6666, #5005, #2255}
@prop _."dequeuer_task" 0 rc
@prop _."queue" {} c
;_.queue = $dict:new()
@prop _."srv_prefixes" {} rc
;_.("srv_prefixes") = {"_xmpp-server._tcp.", "_jabber._tcp."}
@prop _."queued_connections" {waif} ""
@prop _."error_dumps" {} rc
;_.("error_dumps") = {#2}
@prop _."site" "example.com" rc
@prop _."jid_regexp" "^%(%([^@/]+%)@%)?%([A-Za-z0-9%.-]+%.[A-Za-z]+%)%(/%(.*%)%)?$" rc
@prop _."name_cache" {} ""
;_.name_cache = $dict:new()
@prop _."iq" #2255 rc
@prop _."version" "1.0" rc
@prop _."moldable" #5544 rc
;_.("handler") = #225
;_.("is_binary") = 1
;_.("port") = 5269
;_.("aliases") = {"XMPP daemon", "Jabber daemon"}
;_.("description") = "An XMPP daemon that handles server-to-server (s2s) communications. Here it was that Kenny Root was finally driven mad."

@verb _:"find_type_handler" this none this
@program _:find_type_handler
{type} = args;
if (index(type, ":") > 1)
  type = type[1..index(type, ":") - 1];
endif
for handler in (this.all_type_handlers)
  if (handler.stanza == type)
    return handler;
  endif
endfor
return $nothing;
.

@verb _:"new_type_handler" this none this
@program _:new_type_handler
{type} = args;
generic = this:find_type_handler(type);
if (valid(generic))
  return generic:new();
else
  return $nothing;
endif
.

@verb _:"connect" this none this rxd #2
@program _:connect
{tohost, ?outboundhost = this.site} = args;
if ((caller != this) && (!$perm_utils:controls(caller_perms(), this)))
  raise(E_PERM);
endif
for child in (children($local.jabber.handler))
  if ($waif:isa(child.stream, $waif))
    if ($network:is_connected(child))
      "We're already connected. Let the dequeuer work its magic.";
      if (child.stream.to == tohost)
        if ((outboundhost && child.outbound) && (child.stream.from == outboundhost))
          return child;
        elseif (!outboundhost)
          return child;
        endif
      endif
    endif
  endif
endfor
{realhost, realport} = this:find_jabber_hostport(tohost);
if (typeof(conn = open_network_connection(realhost, realport, this)) == ERR)
  return conn;
endif
this.queued_connections[realhost] = {outboundhost, tohost};
"The connection will be opened asynchronously. When the user connects to this daemon, the dequeuer should feed it messages to send.";
suspend(1);
return 1;
"Copied from Jabber daemon (#5269):connect by Kenny (#267) Tue Oct 18 10:26:53 2005 PDT";
"Copied from Jabber daemon (#5269):connect(new) by Xythian (#199) Tue Oct 18 10:41:33 2005 PDT";
.

@verb _:"find_connection_by_id" this none this
@program _:find_connection_by_id
{id} = args;
for child in (children(this.handler))
  if (`child.stream.id == id ! ANY => 0' || `child.stream.id_remote == id ! ANY => 0')
    return child;
  endif
endfor
return $failed_match;
.

@verb _:"find_connection_by_dialback" this none this
@program _:find_connection_by_dialback
{secret} = args;
for child in (children(this.handler))
  dbr = child:get_type_handler("db:result");
  if (dbr.text == secret)
    return child;
  endif
endfor
return $failed_match;
.

@verb _:"clean" this none this
@program _:clean
if ((caller != this) && (!$perm_utils:controls(caller_perms(), this)))
  raise(E_PERM);
endif
for child in (children($local.jabber.handler))
  if (!$network:is_connected(child))
    child:disconnected();
  endif
endfor
.

@verb _:"send" this none this
@program _:send
{object} = args;
if ((caller != this) && (!$perm_utils:controls(caller_perms(), this)))
  raise(E_PERM);
endif
hostpair = {object.from, object.to};
for host in [1..length(hostpair)]
  {user, uhost, resource} = this:split_jid(hostpair[host]);
  hostpair[host] = uhost;
endfor
hostpair = tostr(hostpair[1], ":", hostpair[2]);
try
  this.queue[hostpair] = {@this.queue[hostpair], object};
except (E_RANGE)
  this.queue[hostpair] = {object};
endtry
this:run_dequeuer();
.

@verb _:"dequeuer" this none this
@program _:dequeuer
if (caller != this)
  raise(E_PERM);
endif
while (hostpairs = $dict:keys(this.queue))
  for key in (hostpairs)
    if (idx = index(key, "|inbound"))
      hostpair = key[1..idx - 1];
      outbound = 0;
    else
      hostpair = key;
      outbound = 1;
    endif
    {fromhost, tohost} = $string_utils:explode(hostpair, ":");
    conn = `this:connect(tohost, outbound ? fromhost | "") ! ANY';
    if (typeof(conn) == OBJ)
      for msg in (this.queue[key])
        req_conn_level = `msg.connection_level ! E_PROPNF => 2';
        if (conn.stream:connected(req_conn_level))
          conn:send(msg);
          this.queue[key] = setremove(this.queue[key], msg);
        else
          (msg:reduce_tries() <= 0) && this:failed_queue(key, msg);
        endif
      endfor
    elseif (typeof(conn) == ERR)
      for msg in (this.queue[key])
        (msg:reduce_tries() <= 0) && this:failed_queue(key, msg);
      endfor
    endif
    if (length(this.queue[key]) == 0)
      this.queue = $dict:remove(this.queue, key);
    endif
  endfor
  suspend(60);
endwhile
.

@verb _:"run_dequeuer" this none this rxd #2
@program _:run_dequeuer
if ((caller != this) && (!$perm_utils:controls(caller_perms(), this)))
  raise(E_PERM);
endif
if ($code_utils:task_valid(this.dequeuer_task))
  `resume(this.dequeuer_task) ! E_INVARG, E_PERM';
else
  fork t (0)
    this.dequeuer_task = task_id();
    try
      this:dequeuer();
    except v (ANY)
      this:report_traceback(v);
    endtry
    this.dequeuer_task = 0;
  endfork
  this.dequeuer_task = t;
endif
.

@verb _:"failed_queue" this none this
@program _:failed_queue
{host, msg} = args;
if ((caller != this) && (!$perm_utils:controls(caller_perms(), this)))
  raise(E_PERM);
endif
this.queue[host] = setremove(this.queue[host], msg);
`msg:failure() ! ANY';
.

@verb _:"find_jabber_hostname" this none this
@program _:find_jabber_hostname
{host} = args;
for prefix in (this.srv_prefixes)
  hdr = $local.resolver.header:new();
  hdr:add_query(hdr.class.query:new(tostr(prefix, host), "T_SRV", "C_IN"));
  if (typeof(res = `$local.resolver:send(hdr) ! ANY => 0') == $waif.type)
    if (length(res.answers))
      srv = res.answers[1].rdata;
      return srv[4];
    endif
  endif
endfor
return host;
.

@verb _:"find_outbound_hostname" this none this
@program _:find_outbound_hostname
{target} = args;
if (hosts = `this.queued_connections[target] ! E_RANGE => 0')
  this.queued_connections = $dict:remove(this.queued_connections, target);
  return hosts;
else
  this.owner:debug(tostr("FAILED TO FIND CONNECTION TO ", target, "!"));
  return $network.site;
endif
.

@verb _:"report_traceback" this none this
@program _:report_traceback
{v} = args;
if (caller != this)
  return;
endif
formatted = $code_utils:format_traceback(v[4], v[1], v[2]);
for dude in (this.error_dumps)
  dude:tell("+----JABBER TRACEBACK");
  for line in (formatted)
    dude:tell("| ", line);
  endfor
  dude:Tell("+-----");
endfor
return;
"Copied from generic HTTP/1.0 service handler (#580):report_traceback by Kenny (#2) Thu Sep 22 10:54:08 2005 CDT";
"Copied from Jabber s2s service handler (#606):report_traceback by Kenny (#2) Fri Oct  7 20:23:49 2005 CDT";
.

@verb _:"@resume" this none none
@program _:@resume
this:run_dequeuer();
.

@verb _:"split_jid" this none this
@program _:split_jid
{jid} = args;
if (matches = match(jid, $local.jabber.jid_regexp))
  bugcatcher = substitute("%1", matches);
  user = substitute("%2", matches);
  host = substitute("%3", matches);
  resource = substitute("%5", matches);
  if (!bugcatcher)
    user = "";
  endif
  return {user, host, resource};
else
  raise(E_INVARG, "invalid JID, must be in the format [<user>@]host[/<resource>]");
endif
.

"Skipping #5269:"find_jabber_hostname(old)"...
@verb _:"clear_name_cache" this none this
@program _:clear_name_cache
cur_time = time();
for key in ($dict:keys(this.name_cache))
  if (cur_time > this.name_cache[key][1])
    this.name_cache = $dict:remove(this.name_cache, key);
  endif
endfor
.

@verb _:"find_jabber_hostport" this none this
@program _:find_jabber_hostport
{host} = args;
if (random(3) == 1)
  fork (0)
    this:clear_name_cache();
  endfork
endif
time_now = time();
if (`{expiration, data} = this.name_cache[host] ! E_RANGE => 0')
  if (time_now > expiration)
    this.name_cache = $dict:remove(this.name_cache, host);
  else
    return data;
  endif
endif
tmphost = host;
suffixes = {host};
while (idx = index(tmphost, "."))
  if ((length(tmphost) > 1) && (rindex(tmphost, ".") != idx))
    tmphost = tmphost[idx + 1..$];
    suffixes = {@suffixes, tmphost};
  else
    break;
  endif
endwhile
for suffix in (suffixes)
  for prefix in (this.srv_prefixes)
    hdr = $local.resolver.header:new();
    hdr:add_query(hdr.class.query:new(tostr(prefix, suffix), "T_SRV", "C_IN"));
    if (typeof(res = `$local.resolver:send(hdr) ! ANY => 0') == $waif.type)
      if (length(res.answers))
        srv = res.answers[1].rdata;
        if (typeof(srv) == LIST)
          this.name_cache[host] = {res.answers[1].ttl + time_now, {srv[4], srv[3]}};
          return {srv[4], srv[3]};
        endif
      endif
    endif
  endfor
endfor
for suffix in (suffixes)
  hdr = $local.resolver.header:new();
  hdr:add_query(hdr.class.query:new(suffix, "T_A", "C_IN"));
  if (typeof(res = `$local.resolver:send(hdr) ! ANY => 0') == $waif.type)
    if (length(res.answers))
      this.name_cache[host] = {time_now + 3600, {suffix, this.port}};
      return {suffix, this.port};
    endif
  endif
endfor
this.name_cache[host] = {time_now + 3600, {host, this.port}};
return {host, this.port};
.

@verb _:"connect(new)" this none this
@program _:connect(new)
{tohost, ?outboundhost = this.site} = args;
if ((caller != this) && (!$perm_utils:controls(caller_perms(), this)))
  raise(E_PERM);
endif
for child in (children($local.jabber.handler))
  if ($waif:isa(child.stream, $waif))
    if ($network:is_connected(child))
      "We're already connected. Let the dequeuer work its magic.";
      if (child.stream.to == tohost)
        if ((outboundhost && child.outbound) && (child.stream.from == outboundhost))
          return child;
        elseif (!outboundhost)
          return child;
        endif
      endif
    endif
  endif
endfor
{realhost, realport} = this:find_jabber_hostport(tohost);
if (typeof(conn = open_network_connection(realhost, realport, this)) == ERR)
  return conn;
endif
this.queued_connections[realhost] = {outboundhost, tohost};
"The connection will be opened asynchronously. When the user connects to this daemon, the dequeuer should feed it messages to send.";
suspend(1);
return 1;
"Copied from Jabber daemon (#5269):connect by Kenny (#267) Tue Oct 18 10:26:53 2005 PDT";
.

@verb _:"find_outbound_tofrom" this none this
@program _:find_outbound_tofrom
{target} = args;
if (fromhost = `this.queued_connections[target] ! E_RANGE => 0')
  this.queued_connections = $dict:remove(this.queued_connections, target);
  return fromhost;
else
  this.owner:debug(tostr("FAILED TO FIND CONNECTION TO ", target, "!"));
  return $network.site;
endif
.

@verb _:"version" this none this
@program _:version
return {{"name", "MOO-XMPP"}, {"version", this.version}, {"os", "MOO VM"}};
.

"***finished***
